#!/bin/bash
#-----------------------------------------------------------------------------
#
# nxlaunch-[ssh|scp]
#
# NX Launcher enables you to launch NX Sessions to a remote Terminal Server
# either via SSH X11 forwarding or NXproxy. NX Launcher uses the NX 3 OSS
# components comprising an outstanding X-Window compression technology
# developed by NoMachine.
#
# The `nxlaunch-[ssh|scp]' commands are launching SSH or SCP sessions
# to the host from where `nxlaunch' has been started.
#
#-----------------------------------------------------------------------------
#
# Author: (c) 2004,2025 - Invoca Systems <simon.matter@invoca.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#-----------------------------------------------------------------------------

declare -r VERSION="4.4.10"

# Treat unset variables and parameters other than the special
# parameters `@' and `*' as an error when performing parameter expansion.
set -o nounset

# Print an error message to STDERR and exit
exit_message() {
  local -i RETVAL=$1
  local PREFIX=
  if (( RETVAL )); then
    PREFIX="Error: "
  fi
  shift
  echo "${PREFIX}${*}" >&2
  # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
  exit $RETVAL
}

# Unable to run with BASH in POSIX mode
if shopt -qo posix; then
  exit_message 1 "unable to run with \`bash' in POSIX mode, exiting"
fi

# We really need `$DISPLAY'
if ! (( ${DISPLAY+1} )); then
  exit_message 1 "environment variable [\$DISPLAY] is not set, exiting"
elif ! [[ "$DISPLAY" ]]; then
  exit_message 1 "environment variable [\$DISPLAY] is empty, exiting"
fi

# Detect basename
declare TMP_NAME="$0"
declare TMP_PWD="$PWD"
while [[ -L "$TMP_NAME" ]]; do
  if [[ "$TMP_NAME" != "${TMP_NAME/\/}" ]]; then
    cd "${TMP_NAME%/*}" 2> /dev/null ||
      exit_message 1 "unable to change to directory ${TMP_NAME%/*}, exiting"
  fi
  TMP_NAME="$(readlink "${TMP_NAME##*/}")"
done
cd "$TMP_PWD" 2> /dev/null ||
  exit_message 1 "unable to change to directory ${TMP_PWD}, exiting"
declare -r BASENAME="${TMP_NAME##*/}"
unset TMP_NAME
unset TMP_PWD
declare -r SHORTNAME="${BASENAME%%-*}"

# Variables from `nxlaunch'
declare -r LOCALHOST="localhost"
declare -ri REVBASEPORT=2000
declare -ri REMBASEDISP=1000

# SSH key type [ dsa | rsa | ecdsa | ed25519 ]
declare -r SSHKEYTYPE="rsa"

# Define individual `ssh-keygen' options depending on SSH key type
case "$SSHKEYTYPE" in
  dsa|ecdsa|ed25519)
    declare -r SSHKEYGENOPTS="-t $SSHKEYTYPE"
    ;;
  rsa)
    declare -r SSHKEYGENOPTS="-t $SSHKEYTYPE -b 3072"
    ;;
esac

#-----------------------------------------------------------------------------

# Help message
helpmessage() {
  cat << EOF
$BASENAME version $VERSION
Copyright (c) 2004,2025 - Invoca Systems
This program may be freely redistributed under the terms of the GNU GPL

$BASENAME launches an SSH or SCP session to the host from where
the current X session has been started. It auto detects whether
it is an $SHORTNAME session or a normal local or remote X session.
It automatically replaces the string \`NXLAUNCHHOST' by the correct
hostname to reach the originating host.

$BASENAME accepts almost all ssh(1) options. If it is called as
${SHORTNAME}-scp then it calls scp(1) instead.

Usage: $BASENAME [OPTION]... <[user@]NXLAUNCHHOST> [<option>]...

Special options:
add_key         adds local public key ~/.${SHORTNAME}/id_${SSHKEYTYPE}.pub to
                the remote sides ~/.ssh/authorized_keys

Examples:
$BASENAME add_key user@NXLAUNCHHOST
$BASENAME -C user@NXLAUNCHHOST hostname
$BASENAME -o StrictHostKeyChecking=no -o CheckHostIP=no NXLAUNCHHOST ls
${SHORTNAME}-scp -C -p file.tar.gz user@NXLAUNCHHOST:Desktop/
EOF
}

#-----------------------------------------------------------------------------

if ! (( $# )); then
  helpmessage
  exit 1
fi

# Do we run `ssh' or `scp'?
declare -r SSHBIN="${0##*-}"
if [[ "$SSHBIN" == "ssh" ]]; then
  declare -r PORTOPTION="-p"
else
  declare -r PORTOPTION="-P"
fi

# Find out how to connect
declare DISPLAYNUMBER=
declare SSHHOST=
declare SSHPORTOPT=
DISPLAYNUMBER="${DISPLAY##*:}"
DISPLAYNUMBER="${DISPLAYNUMBER%%.*}"
if (( DISPLAYNUMBER > REMBASEDISP )); then
  # Proxy mode
  SSHHOST="$LOCALHOST"
  SSHPORTOPT="$PORTOPTION $(( REVBASEPORT + DISPLAYNUMBER ))"
elif [[ "${SSH_CONNECTION%% *}" ]]; then
  # SSH X11 mode
  SSHHOST="${SSH_CONNECTION%% *}"
else
  exit_message 1 "don't know how to handle this configuration, exiting"
fi

# Create private keystore
# shellcheck disable=SC2174 # doesn't apply to subdirectory in `$HOME'
mkdir -m 0700 -p ~/."${SHORTNAME}"
if ! [[ -f ~/."${SHORTNAME}"/id_"${SSHKEYTYPE}" ]]; then
  echo "Private/public \`${SSHKEYTYPE}' key pair for [${SHORTNAME}] not found, creating now:"
  echo
  # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
  ssh-keygen $SSHKEYGENOPTS -f ~/."${SHORTNAME}"/id_"${SSHKEYTYPE}" -C "${BASENAME}@$(hostname)" -N ''
  echo
fi

# Add our public key to the remote host
declare -i RETVAL=0
if [[ "$1" == "add_key" ]]; then
  if [[ "$SSHBIN" != "ssh" ]]; then
    exit_message 1 "\`add_key' can only be used in SSH mode, exiting"
  fi
  shift
  if (( $# )); then
    # TODO: This could be done with newer versions of `ssh-copy-id' as well
    # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
    ssh -i ~/."${SHORTNAME}"/id_"${SSHKEYTYPE}" \
      -o NoHostAuthenticationForLocalhost=yes \
      -o UserKnownHostsFile=~/."${SHORTNAME}"/known_hosts \
      $SSHPORTOPT "${@//NXLAUNCHHOST/${SSHHOST}}" \
      mkdir -m 0700 -p "\${HOME}"/.ssh ";" \
      cat ">>" "\${HOME}"/.ssh/authorized_keys ";" \
      chmod 0644 "\${HOME}"/.ssh/authorized_keys < ~/."${SHORTNAME}"/id_"${SSHKEYTYPE}".pub
    RETVAL=$?
  else
    RETVAL=1
  fi
  if (( RETVAL )); then
    helpmessage
  fi
  exit $RETVAL
fi

# Run `ssh' or `scp'
# shellcheck disable=SC2086 # doesn't apply to integer variables or lists
exec "$SSHBIN" -i ~/."${SHORTNAME}"/id_"${SSHKEYTYPE}" \
  -o NoHostAuthenticationForLocalhost=yes \
  -o UserKnownHostsFile=~/."${SHORTNAME}"/known_hosts \
  $SSHPORTOPT "${@//NXLAUNCHHOST/${SSHHOST}}"
